module ReportsMash
  module Engine
    class Engine
      module ClassMethods
        def instance
          @instance ||= self.new
        end
      end
      extend ClassMethods
      
      attr_accessor :delivery_queue
      attr_accessor :pid
      attr_accessor :collector

      def initialize
        @transactions = {}
        @engine_start = Time.now
        @delivery_queue = []
        @db_transactions = {}
        @pid = Process.pid
        @started = false
      end

      def boot_from_fork
        return if @started
        @started = true
        @pid = Process.pid
        @engine_start = Time.now

        $stderr.puts "Initializing plugin ReportsMash in child #{Process.ppid} -> #{Process.pid}\n"
        begin
          start_worker
        rescue => e
          $stderr.puts "Error while starting worker #{e.inspect}"
        end
      end

      def transaction_start(request)
        if @pid != Process.pid && !@started
          # we have been forked. Make sure engine worker is running in this child
          ReportsMash::Engine.boot_from_fork()
        end
        transaction = ReportsMash::Engine::Transaction.new(@instance, request)
        ReportsMash::Engine::ThreadState.get_current_state.transaction = transaction
        request["X-REPORTSMASH-TRANSACTION-ID"] = transaction.tid
        @transactions[transaction.tid] = transaction
      end

      def transaction_end(start_time, stop_time, e, response, env)
        transaction = @transactions[env["X-REPORTSMASH-TRANSACTION-ID"]]
        (status_code, headers, body) = response       
        transaction.end(start_time, stop_time, status_code, headers)
        # don't allow more than 150 transactions in queue
        @delivery_queue.push transaction if @delivery_queue.count < 150
        env.delete("X-REPORTSMASH-TRANSACTION-ID") if env["X-REPORTSMASH-TRANSACTION-ID"]
        ReportsMash::Engine::ThreadState.clear
      end

      def custom_attr(new_params)
        trans = ReportsMash::Engine::ThreadState.get_current_state.transaction
        trans.user_params.merge!(new_params) if new_params.is_a? Array
      end

      def start_worker
        @worker_thread = ReportsMash::Engine::Threading::WorkerThread.new do
          worker_loop = ReportsMash::Engine::WorkerLoop.new
          worker_loop.run
        end
      end

      def save_db_transaction(id, duration, sql, connection_id)
        trans = ReportsMash::Engine::ThreadState.get_current_state.transaction
        return unless trans
        item = {
          duration: duration,
          sql: sql,
          connection_id: connection_id
        }
        trans.db_transactions.push item
        trans.db_duration_ms += duration
      end

      def save_render_transaction(id, duration, path)
        trans = ReportsMash::Engine::ThreadState.get_current_state.transaction
        return unless trans
        item = {
          duration: duration,
          path: path
        }
        trans.render_transactions.push item
        trans.render_duration_ms += duration
      end

      def save_external_http(item)
        trans = ReportsMash::Engine::ThreadState.get_current_state.transaction
        return unless trans
        trans.external_http_requests.push item
        trans.external_http_duration_ms += item[:duration]
      end

      def save_external_mc(op, duration)
        trans = ReportsMash::Engine::ThreadState.get_current_state.transaction
        return unless trans
        h = {op.to_sym => duration}
        trans.memcached_ops.push h
      end

      def save_external_redis(op, duration)
        trans = ReportsMash::Engine::ThreadState.get_current_state.transaction
        return unless trans
        h = {op.to_sym => duration}
        trans.redis_ops.push h
      end

    end
  end
end

